/*

Copyright (c) 2018, Arvid Norberg, Steven Siloti
Copyright (c) 2018, 2020, Arvid Norberg
All rights reserved.

You may use, distribute and modify this code under the terms of the BSD license,
see LICENSE file.
*/

#ifndef TORRENT_WINDOWS_HPP_INCLUDED
#define TORRENT_WINDOWS_HPP_INCLUDED

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#ifndef VC_EXTRALEAN
#define VC_EXTRALEAN
#endif
#ifndef STRICT
#define STRICT
#endif
#include <windows.h>

#endif // TORRENT_WINDOWS_HPP_INCLUDED



